# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


"""
This module contains entry related tools. Essentially, entries are containers
for calculated information, which is used in many analyses.
"""
